module Comparable exposing
    ( From, default
    , lt, gt, lte, gte, max, min
    , maximum, minimum, sort
    )

{-| This module contains replacements for functions that are imported by default.


# Comparable

@docs From, default


# Basics

Basic functions with explicit comparison.

@docs lt, gt, lte, gte, max, min


# List

List functions with explicit comparison.

@docs maximum, minimum, sort

-}

import Basics
import List


{-| A function to implement comparison for a given type of value.
-}
type alias From value =
    value -> value -> Basics.Order


{-| The default implementation for comparison of values in Elm.
This function implements equality as well as ordering for some types of values - Custom types aren't supported.
-}
default : From comparable
default =
    Basics.compare



-- Basics


{-| is the first value less than the second?
-}
lt : From value -> value -> value -> Bool
lt order first second =
    order first second == Basics.LT


{-| is the first value greater than the second?
-}
gt : From value -> value -> value -> Bool
gt order first second =
    order first second == Basics.GT


{-| is the first value less than or equal to the second?
-}
lte : From value -> value -> value -> Bool
lte order first second =
    order first second /= Basics.GT


{-| is the first value greater than or equal to the second?
-}
gte : From value -> value -> value -> Bool
gte order first second =
    order first second /= Basics.LT


{-| return the greater of two values.
-}
max : From value -> value -> value -> value
max order first second =
    case order first second of
        Basics.GT ->
            first

        _ ->
            second


{-| return the lesser of two values.
-}
min : From value -> value -> value -> value
min order first second =
    case order first second of
        Basics.LT ->
            first

        _ ->
            second



-- List


{-| Find the maximum element in a non-empty list.
-}
maximum : From value -> List value -> Maybe value
maximum order list =
    case list of
        x :: xs ->
            Just (List.foldl (max order) x xs)

        _ ->
            Nothing


{-| Find the minimum element in a non-empty list.
-}
minimum : From value -> List value -> Maybe value
minimum order list =
    case list of
        x :: xs ->
            Just (List.foldl (min order) x xs)

        _ ->
            Nothing


{-| Sort values from lowest to highest.
-}
sort : From value -> List value -> List value
sort order xs =
    List.sortWith order xs
