module Comparable.Set exposing
    ( Set
    , empty, singleton, insert, remove
    , isEmpty, member, size
    , union, intersect, diff
    , toList, fromList
    , map, foldl, foldr, filter, partition
    )

{-| A set of unique values. The values can be any type.


# Sets

@docs Set


# Build

@docs empty, singleton, insert, remove


# Query

@docs isEmpty, member, size


# Combine

@docs union, intersect, diff


# Lists

@docs toList, fromList


# Transform

@docs map, foldl, foldr, filter, partition

-}

import Comparable
import Comparable.Dict as Dict exposing (Dict)


{-| Represents a set of unique values.
-}
type Set value
    = Set (Dict value ())


{-| Create an empty set.
-}
empty : Set value
empty =
    Set Dict.empty


{-| Create a set with one value.
-}
singleton : value -> Set value
singleton key =
    Set (Dict.singleton key ())


{-| Insert a value into a set.
-}
insert : Comparable.From value -> value -> Set value -> Set value
insert compare key (Set dict) =
    Set (Dict.insert compare key () dict)


{-| Remove a value from a set. If the value is not found, no changes are made.
-}
remove : Comparable.From value -> value -> Set value -> Set value
remove compare key (Set dict) =
    Set (Dict.remove compare key dict)


{-| Determine if a set is empty.
-}
isEmpty : Set value -> Bool
isEmpty (Set dict) =
    Dict.isEmpty dict


{-| Determine if a value is in a set.
-}
member : Comparable.From value -> value -> Set value -> Bool
member compare key (Set dict) =
    Dict.member compare key dict


{-| Determine the number of elements in a set.
-}
size : Set value -> Int
size (Set dict) =
    Dict.size dict


{-| Get the union of two sets. Keep all values.
-}
union : Comparable.From value -> Set value -> Set value -> Set value
union compare (Set firstDict) (Set secondDict) =
    Set (Dict.union compare firstDict secondDict)


{-| Get the intersection of two sets. Keeps values that appear in both sets.
-}
intersect : Comparable.From value -> Set value -> Set value -> Set value
intersect compare (Set firstDict) (Set secondDict) =
    Set (Dict.intersect compare firstDict secondDict)


{-| Get the difference between the first set and the second. Keeps values
that do not appear in the second set.
-}
diff : Comparable.From value -> Set value -> Set value -> Set value
diff compare (Set firstDict) (Set secondDict) =
    Set (Dict.diff compare firstDict secondDict)


{-| Convert a set into a list, sorted from lowest to highest.
-}
toList : Set value -> List value
toList (Set dict) =
    Dict.keys dict


{-| Convert a list into a set, removing any duplicates.
-}
fromList : Comparable.From value -> List value -> Set value
fromList compare list =
    List.foldl (insert compare) empty list


{-| Fold over the values in a set, in compare from lowest to highest.
-}
foldl : (value -> accumulator -> accumulator) -> accumulator -> Set value -> accumulator
foldl func accumulator (Set dict) =
    Dict.foldl (\key _ state -> func key state) accumulator dict


{-| Fold over the values in a set, in compare from highest to lowest.
-}
foldr : (value -> accumulator -> accumulator) -> accumulator -> Set value -> accumulator
foldr func accumulator (Set dict) =
    Dict.foldr (\key _ state -> func key state) accumulator dict


{-| Map a function onto a set, creating a new set with no duplicates.
-}
map : Comparable.From otherValue -> (value -> otherValue) -> Set value -> Set otherValue
map compare toOther set =
    fromList compare (foldl (\x xs -> toOther x :: xs) [] set)


{-| Only keep elements that pass the given test.
-}
filter : Comparable.From value -> (value -> Bool) -> Set value -> Set value
filter compare isKept (Set dict) =
    Set (Dict.filter compare (\key _ -> isKept key) dict)


{-| Create two new sets. The first contains all the elements that passed the
given test, and the second contains all the elements that did not.
-}
partition : Comparable.From value -> (value -> Bool) -> Set value -> ( Set value, Set value )
partition compare isLeft (Set dict) =
    let
        ( firstDict, secondDict ) =
            Dict.partition compare (\key _ -> isLeft key) dict
    in
    ( Set firstDict, Set secondDict )
